clear; close all; clc

set(0,'DefaultFigureWindowStyle','docked')

%% Set Parameters
M = 0.5;
m = 0.2;
l = 0.3;
I = 0.006;
b = 0.2;
g = 9.81;

% Create continuous-time dynamics
Ac = [0 1 0 0;
    0 b*(m*l^2 - I)/(I*(M+m)-M*m*l^2) -m^2*l^2*g/(I*(M+m)-M*m*l^2) 0;
    0 0 0 1;
    0 -b*m*l/(I*(M+m)-M*m*l^2) m*g*l*(M+m)/(I*(M+m)-M*m*l^2) 0];

Bc = [0;
    (I-m*l^2)/(I*(M+m)-M*m*l^2);
    0;
    m*l/(I*(M+m)-M*m*l^2)];

%% Convert to Discrete-time
Ts = 0.1;

ABdisc = expm([Ac Bc; zeros(1,5)]*Ts);
A = ABdisc(1:4,1:4);
B = ABdisc(1:4,end);

%% Set the initial condition and target

x0 = [2;0;0;0];
xstar = [10;0;0;0];

%% MPC

% set new cost function weights
Q = eye(4);
R = 1;

% set the prediction horizon in seconds
N = 1;
N = N/Ts; % convert to discrete-time steps

% Create Prediction Horizon Model and Cost
Delta = A;
Phitemp = B;
Phi = B;
Qbar = Q;
Rbar = R;
for ii = 2:N
    Delta = [A; Delta*A];
    Phitemp = [A*Phitemp B];
    Phi = [Phi zeros(4*(ii-1),1); Phitemp];
    Qbar = blkdiag(Qbar, Q);
    Rbar = blkdiag(Rbar, R);
end



% Fixed Hession
H = 2 * Phi.'*Qbar*Phi + Rbar;

% Set the equality constraints to null
Aeq = zeros(0,length(H));
Beq = zeros(0,1);

% set the MPC Iterations to a large number
opt = mpcActiveSetOptions;
opt.MaxIterations = 10000;

% Set the constraints
xmin = [-5;-3;-0.1;-10];
xmax = [10;3;0.1;10];
umin = -100;
umax = 100;

% propagate the constant target reference and constraints across the horizon
xstarbig = kron(ones(N,1),xstar);
xmin = kron(ones(N,1),xmin);
xmax = kron(ones(N,1),xmax);
umin = kron(ones(N,1),umin);
umax = kron(ones(N,1),umax);

% Create the inequality constraint matrix
Ain = [-Phi;Phi;-eye(N);eye(N)];

% set the initial constraints to 'inactive'
iA = false(size(Ain,1),1);

% set the simulation time
TfinalMPC = 20/Ts;

% preallocate
xMPC = zeros(4,TfinalMPC);
xMPC(:,1) = x0;
uMPCfull = zeros(1,TfinalMPC);

tic
for kk = 1:TfinalMPC-1
    
    % create latest inequality constraint vector
    Bin = [Delta*xMPC(:,kk) - xmin;
        xmax - Delta*xMPC(:,kk);
        -umin;umax];
    
    % create 'Multiplier of the objective function linear term'
    fT = 2*(Phi).'*Qbar*(Delta*xMPC(:,kk) - xstarbig);

    % SOLVE MPC!
    [uMPC, exitflag, iA, lambda] = mpcActiveSetSolver...
        (H, fT, Ain, Bin, Aeq, Beq, iA, opt);
    
    if exitflag == -2 || exitflag == -1 || exitflag == 0
        error(['MPC exitflag = ' num2str(exitflag)])
    end
    
    % extract the latest control action
    uMPCfull(kk) = uMPC(1);
    % propagate system
    xMPC(:,kk+1) = A*xMPC(:,kk) + B*uMPCfull(kk);
end
toc

% Plot
figure
subplot(3,2,1)
plot(0:Ts:TfinalMPC*Ts-Ts,xMPC(1,:)); ylabel('Position')
subplot(3,2,2)
plot(0:Ts:TfinalMPC*Ts-Ts,xMPC(2,:)); ylabel('Velocity')
subplot(3,2,3)
plot(0:Ts:TfinalMPC*Ts-Ts,xMPC(3,:)); ylabel('Angle')
subplot(3,2,4)
plot(0:Ts:TfinalMPC*Ts-Ts,xMPC(4,:)); ylabel('Angle rate')
subplot(3,2,5)
plot(0:Ts:TfinalMPC*Ts-Ts,uMPCfull); ylabel('Control action')

