import numpy as np
import matplotlib.pyplot as plt
import scipy
from scipy.linalg import expm

import control as ct
import control.optimal as obc

import time

import platform
print('Platform Info: ' + str(platform.system_alias(platform.system(),platform.release(),platform.version())))
print('Platform Info: ' + platform.platform(aliased=0, terse=0))
print('Python Version: ' + platform.python_version())
print('Numpy Version: ' + np.__version__)
print('Scipy Version: ' + scipy.__version__)
print('Control Version: ' + ct.__version__)


def convertToDiscreteTime(Ac,Bc,Ts):
    AB = expm(np.concatenate((np.concatenate((Ac, Bc),axis=1),np.zeros((1,5))),axis=0) * Ts)
    A = np.array(AB[0:4,0:4])
    B = np.transpose(np.array([AB[0:4,-1]]))
    return A, B


M = 0.5
m = 0.2
l = 0.3
I = 0.006
b = 0.2
g = 9.81


Ac = np.array([[0, 1, 0, 0],
    [0, b*(m*l*l - I)/(I*(M+m)-M*m*l*l), -m*m*l*l*g/(I*(M+m)-M*m*l*l), 0],
    [0, 0, 0, 1],
    [0, -b*m*l/(I*(M+m)-M*m*l*l), m*g*l*(M+m)/(I*(M+m)-M*m*l*l), 0]])

Bc = np.array([[0],
    [(I-m*l*l)/(I*(M+m)-M*m*l*l)],
    [0],
    [m*l/(I*(M+m)-M*m*l*l)]])


# sampling rate
Ts = 0.1

# convert to discrete time
A, B = convertToDiscreteTime(Ac,Bc,Ts)

# grab number of states
numState = len(A)


# Specify Q and R matrices 
Q = np.eye(numState)
R = np.eye(1)*1

# initial conditions
x_0 = np.array([[2],[0],[0],[0]])

# target state reference
xstar = np.array([[10],[0],[0],[0]])


# include state constraints, set min max velocity and acceleration, assume 
maxCranePosition = 10 # m
minCranePosition = -5 # m
maxVelocity = 3 # m/s
minVelocity = -maxVelocity # m/s
maxCraneAngle = 0.1 # rad
minCraneAngle = -maxCraneAngle # rad
maxCraneAngledot = 10
minCraneAngledot = -maxCraneAngledot

maxForce = 100 # N
minForce = -maxForce # N


# Choose the finite horizon
finiteHorizon = 1 # seconds
# convert to discrete-time steps
finiteHorizon = finiteHorizon/Ts



# create system model (used to mimic the system)
system = ct.ss2io(ct.ss(A, B, np.eye(numState), 0, Ts))

# Assume full state knowledge
C = np.eye(numState)
# create mathematical model that will be the basis of the controller design
model = ct.ss2io(ct.ss(A, B, C, 0, Ts))

# create constraint vector
constraints = [obc.state_range_constraint(system, [minCranePosition, minVelocity, minCraneAngle, minCraneAngledot], [maxCranePosition, maxVelocity, maxCraneAngle, maxCraneAngledot]), obc.input_range_constraint(system, [minForce], [maxForce])]

# create stage cost function on the model, with given desired error signal and zero control input
cost = obc.quadratic_cost(model, Q, R, x0=xstar.flatten(), u0=0)
# note: this uses the 'model' (which could be different to the system)

# online MPC controller object is constructed with finite horizon, cost and constraints
ctrl = obc.create_mpc_iosystem(model, np.arange(0, finiteHorizon) * Ts, cost, constraints)

# create closed-loop feedback 'loop' object using the feedback function to link the controller to the system
loop = ct.feedback(system, ctrl, 1)



res = obc.solve_ocp(model, np.arange(0, finiteHorizon) * Ts, x_0.flatten(), cost, constraints, squeeze=True)

print(res)


# set length of simulation
Nsim = 20
# convert to number of discrete-time steps
Nsim = Nsim/Ts

# Simulate the closed-loop MPC
start = time.time()
tout, xout = ct.input_output_response(loop, np.arange(0, Nsim) * Ts, 0, x_0.flatten())
end = time.time()
print("Computation time = %g seconds" % (end-start))

figlabels = ['Crane Cart Position [m]','Cart Velocity [m/s]','Object Degree from Centre [rad]','Object Degree rate [rad/s]']

for i, y in enumerate(xout):
    plt.figure()
    plt.plot(tout, y)
    plt.plot(tout, xstar[i] * np.ones(tout.shape), 'k--')
    plt.ylabel(figlabels[i])
    plt.xlabel('Time [sec]')
    plt.savefig('fig' + str(i) + '.pdf')



